# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Josephson Junction (Manhattan-Style)
REFERENCE: I.M. Pop et al., Nature 508, 369-372 (2014)
"""
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core.base import QComponent


class jj_manhattan(QComponent):
    """
    The base "JJ_Manhattan" inherits the "QComponent" class.
    
    NOTE TO USER: Please be aware that when designing with this
    qcomponent, one must take care in accounting for the junction
    qgeometry when exporting to to GDS and/or to a simulator. This
    qcomponent should not be rendered for EM simulation.

    This creates a "Manhattan"-style Josephson Junction consisting
    of two overlapping thin metal strips, each connected to a
    larger metallic pad region.

    .. image::
        JJManhattan.png

    .. meta::
        :description: Josephson Junction Manhattan

    Default Options:
        * JJ_pad_lower_width: '4um' -- width of lower JJ metal region
        * JJ_pad_lower_height: '2um' -- height of lower JJ metal region
        * JJ_pad_lower_pos_x: '0' -- the initial x-coord of the lower rectangle
        * JJ_pad_lower_pos_y: '0' -- the initial y-coord of the lower rectangle
        * finger_lower_width: '1um' -- the width of the overlapping rectangular finger(s)
        * finger_lower_height: '20um' -- the length of the overlapping rectangular finger(s)
        * extension: '1um' -- the length of the fingers extending beyond the cross-point
    """
    # Default drawing options
    default_options = Dict(JJ_pad_lower_width='25um',
                           JJ_pad_lower_height='10um',
                           JJ_pad_lower_pos_x='0',
                           JJ_pad_lower_pos_y='0',
                           finger_lower_width='1um',
                           finger_lower_height='20um',
                           extension='1um')
    """Default drawing options"""

    # Name prefix of component, if user doesn't provide name
    component_metadata = Dict(short_name='component')
    """Component metadata"""

    def make(self):
        """Convert self.options into QGeometry."""

        p = self.parse_options()  # Parse the string options into numbers

        # draw the lower pad as a rectangle
        JJ_pad_lower = draw.rectangle(p.JJ_pad_lower_width,
                                      p.JJ_pad_lower_height,
                                      p.JJ_pad_lower_pos_x,
                                      p.JJ_pad_lower_pos_y)

        finger_lower = draw.rectangle(
            p.finger_lower_width, p.finger_lower_height, p.JJ_pad_lower_pos_x,
            0.5 * (p.JJ_pad_lower_height + p.finger_lower_height))

        # fudge factor to merge the two options
        finger_lower = draw.translate(finger_lower, 0.0, -0.0001)

        # merge the lower pad and the finger into a single object
        design = draw.union(JJ_pad_lower, finger_lower)

        # copy the pad/finger and rotate it by 90 degrees
        design2 = draw.rotate(design, 90.0)

        # translate the second pad/finger to achieve the desired extension
        design2 = draw.translate(
            design2, 0.5 * (p.JJ_pad_lower_height + p.finger_lower_height) -
            0.5 * p.finger_lower_width - p.extension,
            0.5 * (p.JJ_pad_lower_height + p.finger_lower_height) -
            0.5 * p.finger_lower_width - p.extension)

        final_design = draw.union(design, design2)

        # translate the final design so that the bottom left
        # corner of the lower pad is at the origin
        final_design = draw.translate(final_design, 0.5 * p.JJ_pad_lower_width,
                                      0.5 * p.JJ_pad_lower_height)

        # now translate so that the design is centered on the
        # user-defined coordinates (pos_x, pos_y)
        final_design = draw.translate(final_design, p.pos_x, p.pos_y)

        geom = {'design': final_design}
        self.add_qgeometry('poly', geom, layer=p.layer, subtract=False)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from pyEPR.hfss import parse_units as __parse_units_hfss__
from pyEPR.hfss import \
    unparse_units  # not used here, but in imports of this file

# See also: is_variable_name, is_numeric_possible
#from ... import Dict
from ...toolbox_metal.parsing import parse_value

__all__ = ['parse_value_hfss', 'unparse_units']


def parse_value_hfss(*args):
    """Parse to HFSS units (from user units)."""
    return __parse_units_hfss__(*args)


#TODO: function to itterate and convert user units to


def to_ansys_units(
        value):  # can make more efifiecnt if we assume this is already a float
    """Converve given value to ansys units.

    Args:
        value (float): Value
    """
    __parse_units_hfss__(value)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This is the CircleCaterpillar module."""

from shapely.geometry import CAP_STYLE, JOIN_STYLE
from qiskit_metal import draw  # , Dict
from qiskit_metal.qlibrary.core import QComponent
#from qiskit_metal import is_true


class CircleCaterpillar(QComponent):
    """A single configurable circle.

    Inherits QComponent class.

    .. image::
        CircleCaterpillar.png

    .. meta::
        :description: Circle Caterpillar

    Default Options:
        * segments: '5'
        * distance: '1.2'
        * radius: '300um'
        * resolution: '16'
        * cap_style: 'round' -- Valid options are 'round', 'flat', 'square'
        * subtract: 'False'
        * helper: 'False'
    """

    default_options = dict(
        segments='5',
        distance='1.2',
        radius='300um',
        resolution='16',
        cap_style='round',  # round, flat, square
        # join_style = 'round', # round, mitre, bevel
        # General
        subtract='False',
        helper='False')
    """Default drawing options"""

    TOOLTIP = """A single configurable circle"""

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc."""
        p = self.p  # p for parsed parameters. Access to the parsed options.

        # create the geometry
        caterpillar = [
            draw.Point(p.pos_x - p.radius * i * p.distance, p.pos_y).buffer(
                p.radius,
                resolution=int(p.resolution),
                cap_style=getattr(CAP_STYLE, p.cap_style),
                #join_style = getattr(JOIN_STYLE, p.join_style)
            ) for i in range(int(p.segments))
        ]
        caterpillar = draw.union(caterpillar)

        poly = draw.Polygon([(0, 0), (0.5, 0), (0.25, 0.5)])
        poly = draw.translate(poly, p.pos_x, p.pos_y)
        poly = draw.rotate(poly, angle=65)
        caterpillar = draw.subtract(caterpillar, poly)

        # rect = draw.rectangle(p.radius*0.75, p.radius*0.23,
        #                      xoff=p.pos_x+p.radius*0.3,
        #                      yoff=p.pos_y+p.radius*0.4)
        #caterpillar = draw.subtract(caterpillar, rect)
        # print(caterpillar)

        # add qgeometry
        #self.add_qgeometry('poly', {'mount': rect})
        self.add_qgeometry('poly', {'caterpillar': caterpillar})
        # subtract=p.subtract,
        #                   helper=p.helper, layer=p.layer, chip=p.chip)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from copy import deepcopy
from qiskit_metal.toolbox_python.attr_dict import Dict
from .base import QComponent
import pprint


class BaseQubit(QComponent):
    """Qubit base class. Use to subscript, not to generate directly.

    Has connection lines that can be added

    Inherits components.QComponent class

    Default Options:
        * pos_x: '0um'
        * pos_y: '0um'
        * connection_pads: empty Dict -- The dictionary which contains all active connection lines for the qubit.
        * _default_connection_pads: empty Dict -- The default values for the (if any) connection lines of the qubit.
    """

    default_options = Dict(connection_pads=Dict(),
                           _default_connection_pads=Dict())
    """Default options."""

    component_metadata = Dict(short_name='Q', _qgeometry_table_poly='True')
    """Component metadata"""

    TOOLTIP = """Qubit"""

    def __init__(self,
                 design,
                 name: str = None,
                 options: Dict = None,
                 options_connection_pads: dict = None,
                 make: bool = True,
                 **kwargs):
        """
        Args:
            design (QDesign): The parent design.
            name (str): Name of the component.
            options (dict): User options that will override the defaults.  Defaults to None.
            options_connection_pads(dict): User options for connection pads on qubit
            make (bool): True if the make function should be called at the end of the init.
                    Options be used in the make function to create the geometry.  Defaults to None.
        """
        super().__init__(design, name, options=options, make=False)

        if self.status == 'Not Built':
            # Component is not registered in design.
            # This qubit was not added to design.
            # self.logger.warning(
            # 'In BaseQubit.__init(), the qubit has not been added to design. The component is exiting with None.')
            return None
        if options_connection_pads:
            self.options.connection_pads.update(options_connection_pads)

        try:
            self._set_options_connection_pads()
        except Exception as e:
            raise Exception(
                f"Unable to create connection pads using given parameters: {options_connection_pads}.\n\n If given parameters is None, check to make sure you don't have any invalid child parameters in your connection_pads parameter.\n\n If you don't want any pads, ensure neither options_connection_pads nor options[connection_pads] are parameters\n\n Exception is: {e} "
            )
        if make:
            self.rebuild()

    def _set_options_connection_pads(self):
        """Applies the default options."""
        # class_name = type(self).__name__
        assert '_default_connection_pads' in self.design.template_options[
            self.class_name], f"""When
        you define your custom qubit class please add a _default_connection_pads
        dictionary name as default_options['_default_connection_pads']. This should specify the default
        creation options for the connection. """

        # Not sure if it best to remove it from options to keep
        del self.options._default_connection_pads
        # the self.options cleaner or not, since the options currently copies in the template. This is
        # potential source of bugs in the future
        for name in self.options.connection_pads:
            my_options_connection_pads = self.options.connection_pads[name]
            self.options.connection_pads[name] = deepcopy(
                self.design.template_options[
                    self.class_name]['_default_connection_pads'])
            self.options.connection_pads[name].update(
                my_options_connection_pads)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name

from PySide6 import QtCore
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtGui import QBrush, QColor, QFont, QIcon

# from ...utility._handle_qt_messages import slot_catch_error
from ...utility._toolbox_qt import blend_colors
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .table_view_all_components import QTableView_AllComponents


class QTableModel_AllComponents(QAbstractTableModel):
    """Design components Table model that shows the names of the components and
    their class names etc.

    This class extends the `QAbstractTableModel` class.

    MVC class
    See https://doc.qt.io/qt-5/qabstracttablemodel.html

    Can be accessed with
        t = gui.ui.tableComponents
        model = t.model()
        index = model.index(1,0)
        model.data(index)
    """
    __timer_interval = 500  # ms

    def __init__(self,
                 gui,
                 logger,
                 parent=None,
                 tableView: 'QTableView_AllComponents' = None):
        """
        Args:
            gui (MetalGUI): The GUI.  Defaults to None.
            logger (logger): The logger.
            parent (QWidget): Parent widget.  Defaults to None.
            tableView (QTableView_AllComponents): The table view.  Defaults to None.
        """
        super().__init__(parent=parent)
        self.logger = logger
        self.gui = gui
        self._tableView = tableView  # the view used to preview this model, used to refresh
        self.columns = [
            'Name', 'QComponent class', 'QComponent module', 'Build status',
            'id'
        ]
        self._row_count = -1

        self._create_timer()

    @property
    def design(self):
        """Returns the design."""
        return self.gui.design

    def _create_timer(self):
        """Refresh the model number of rows, etc.

        there must be a smarter way?
        """
        self._timer = QtCore.QTimer(self)
        self._timer.start(self.__timer_interval)
        self._timer.timeout.connect(self.refresh_auto)

    def refresh(self):
        """Force refresh.

        Completely rebuild the model.
        """
        self.beginResetModel()
        try:
            self._row_count = self.rowCount()
        finally:
            self.endResetModel()

    def refresh_auto(self):
        """Automatic refresh, update row count, view, etc."""
        # We could not do if the widget is hidden
        new_count = self.rowCount()

        # if the number of rows have changed
        if self._row_count != new_count:
            # self.logger.info('Number of components changed')

            # Wrap the reset logic in beginResetModel and endResetModel
            self.beginResetModel()
            try:

                # When a model is reset it should be considered that all
                # information previously retrieved from it is invalid.
                # This includes but is not limited to the rowCount() and
                # columnCount(), flags(), data retrieved through data(), and roleNames().
                # This will loose the current selection.
                # self.modelReset.emit()

                self._row_count = new_count
            finally:
                self.endResetModel()

            # for some reason the horizontal header is hidden even if i call this in init
            self._tableView.horizontalHeader().show()
            self.update_view()

    def update_view(self):
        """Updates the view."""
        if self._tableView:
            self._tableView.resizeColumnsToContents()

    def rowCount(self, parent: QModelIndex = None):  # pylint: disable=unused-argument
        """Returns the number of rows.

        Args:
            parent (QModelIndex): Unused.  Defaults to None.

        Returns:
            int: The number of rows
        """
        if self.design:  # should we just enforce this
            num = int(len(self.design.components))
            if num == 0:
                self._tableView.show_placeholder_text()
            else:
                self._tableView.hide_placeholder_text()
            return num
        else:
            self._tableView.show_placeholder_text()
            return 0

    def columnCount(self, parent: QModelIndex = None):  # pylint: disable=unused-argument
        """Returns the number of columns.

        Args:
            parent (QModelIndex): Unused.  Defaults to None.

        Returns:
            int: The number of columns
        """
        return len(self.columns)

    def headerData(self,
                   section,
                   orientation: Qt.Orientation,
                   role=Qt.DisplayRole):
        """Set the headers to be displayed.

        Args:
            section (int): Section number
            orientation (Qt orientation): Section orientation
            role (Qt display role): Display role.  Defaults to DisplayRole.

        Returns:
            str: The header data, or None if not found
        """

        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if section < len(self.columns):
                    # print(self.columns[section])
                    return self.columns[section]

        elif role == Qt.FontRole:
            if section == 0:
                font = QFont()
                font.setBold(True)
                return font

    def flags(self, index):
        """Set the item flags at the given index.

        Args:
            index (QModelIndex): The index

        Returns:
            Qt flags: Flags from Qt
        """
        # https://doc.qt.io/qt-5/qt.html#ItemFlag-enum

        if not index.isValid():
            return Qt.ItemIsEnabled

        return Qt.ItemFlags(
            QAbstractTableModel.flags(self, index) |
            Qt.ItemIsSelectable)  # | Qt.ToolTip)  # ItemIsEditable

    # @slot_catch_error()
    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        """Depending on the index and role given, return data. If not returning
        data, return None (PySide equivalent of QT's "invalid QVariant").

        Returns:
            str: Data depending on the index and role
        """

        if not index.isValid() or not self.design:
            return

        component_name = list(self.design.components.keys())[index.row()]

        if role == Qt.DisplayRole:

            if index.column() == 0:
                return str(component_name)
            elif index.column() == 1:
                return str(
                    self.design.components[component_name].__class__.__name__)
            elif index.column() == 2:
                return str(
                    self.design.components[component_name].__class__.__module__)
            elif index.column() == 3:
                return str(self.design.components[component_name].status)
            elif index.column() == 4:
                return str(self.design.components[component_name].id)

        # The font used for items rendered with the default delegate. (QFont)
        elif role == Qt.FontRole:
            if index.column() == 0:
                font = QFont()
                font.setBold(True)
                return font

        elif role == Qt.BackgroundRole:

            component = self.design.components[component_name]
            if component.status != 'good':  # Did the component fail the build
                #    and index.column()==0:
                if not self._tableView:
                    return QBrush(QColor('#FF0000'))
                table = self._tableView
                color = table.palette().color(table.backgroundRole())
                color = blend_colors(color, QColor('#FF0000'), r=0.6)
                return QBrush(color)

        elif role == Qt.DecorationRole:

            if index.column() == 0:
                component = self.design.components[component_name]
                if component.status != 'good':  # Did the component fail the build
                    return QIcon(":/sample_shapes/warning")

        elif role == Qt.ToolTipRole or role == Qt.StatusTipRole:
            component = self.design.components[component_name]
            text = f"""Component name= "{component.name}" instance of """\
                f"""class "{component.__class__.__name__}" from module""" \
                    f"""{component.__class__.__module__}" """
            return text



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './plot_window_ui.ui',
# licensing of './plot_window_ui.ui' applies.
#
# Created: Sat Jun 19 22:02:29 2021
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainWindowPlot(object):

    def setupUi(self, MainWindowPlot):
        MainWindowPlot.setObjectName("MainWindowPlot")
        MainWindowPlot.resize(800, 600)
        MainWindowPlot.setIconSize(QtCore.QSize(24, 24))
        self.centralwidget = QtWidgets.QWidget(MainWindowPlot)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        MainWindowPlot.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setGeometry(QtCore.QRect(0, 0, 800, 22))
        self.menubar.setObjectName("menubar")
        MainWindowPlot.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindowPlot)
        self.statusbar.setEnabled(True)
        self.statusbar.setObjectName("statusbar")
        MainWindowPlot.setStatusBar(self.statusbar)
        self.toolBar = QToolBarExpanding(MainWindowPlot)
        self.toolBar.setIconSize(QtCore.QSize(20, 20))
        self.toolBar.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        self.toolBar.setObjectName("toolBar")
        MainWindowPlot.addToolBar(QtCore.Qt.TopToolBarArea, self.toolBar)
        self.actionPan = QtGui.QAction(MainWindowPlot)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/plot/pan"), QtGui.QIcon.Normal,
                       QtGui.QIcon.On)
        self.actionPan.setIcon(icon)
        self.actionPan.setObjectName("actionPan")
        self.actionZoom = QtGui.QAction(MainWindowPlot)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/plot/zoom"), QtGui.QIcon.Normal,
                        QtGui.QIcon.Off)
        self.actionZoom.setIcon(icon1)
        self.actionZoom.setObjectName("actionZoom")
        self.actionConnectors = QtGui.QAction(MainWindowPlot)
        self.actionConnectors.setCheckable(True)
        self.actionConnectors.setChecked(False)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/connectors"), QtGui.QIcon.Normal,
                        QtGui.QIcon.Off)
        self.actionConnectors.setIcon(icon2)
        self.actionConnectors.setObjectName("actionConnectors")
        self.actionCoords = QtGui.QAction(MainWindowPlot)
        self.actionCoords.setCheckable(True)
        self.actionCoords.setChecked(True)
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/plot/point"), QtGui.QIcon.Normal,
                        QtGui.QIcon.Off)
        self.actionCoords.setIcon(icon3)
        self.actionCoords.setObjectName("actionCoords")
        self.actionAuto = QtGui.QAction(MainWindowPlot)
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/plot/autozoom"), QtGui.QIcon.Normal,
                        QtGui.QIcon.Off)
        self.actionAuto.setIcon(icon4)
        self.actionAuto.setObjectName("actionAuto")
        self.actionReplot = QtGui.QAction(MainWindowPlot)
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/plot/refresh_plot"),
                        QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionReplot.setIcon(icon5)
        self.actionReplot.setObjectName("actionReplot")
        self.actionRuler = QtGui.QAction(MainWindowPlot)
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/plot/ruler"), QtGui.QIcon.Normal,
                        QtGui.QIcon.Off)
        self.actionRuler.setIcon(icon6)
        self.actionRuler.setObjectName("actionRuler")
        self.toolBar.addAction(self.actionPan)
        self.toolBar.addAction(self.actionAuto)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionCoords)
        self.toolBar.addAction(self.actionConnectors)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionRuler)

        self.retranslateUi(MainWindowPlot)
        QtCore.QObject.connect(self.actionAuto, QtCore.SIGNAL("triggered()"),
                               MainWindowPlot.auto_scale)
        QtCore.QObject.connect(self.actionConnectors,
                               QtCore.SIGNAL("triggered(bool)"),
                               MainWindowPlot.set_show_pins)
        QtCore.QObject.connect(self.actionCoords,
                               QtCore.SIGNAL("triggered(bool)"),
                               MainWindowPlot.set_position_track)
        QtCore.QObject.connect(self.actionPan, QtCore.SIGNAL("triggered()"),
                               MainWindowPlot.pan)
        QtCore.QObject.connect(self.actionZoom, QtCore.SIGNAL("triggered()"),
                               MainWindowPlot.zoom)
        QtCore.QObject.connect(self.actionReplot, QtCore.SIGNAL("triggered()"),
                               MainWindowPlot.replot)
        QtCore.QMetaObject.connectSlotsByName(MainWindowPlot)

    def retranslateUi(self, MainWindowPlot):
        MainWindowPlot.setWindowTitle(
            QtWidgets.QApplication.translate("MainWindowPlot", "MainWindow",
                                             None, -1))
        self.toolBar.setWindowTitle(
            QtWidgets.QApplication.translate("MainWindowPlot", "toolBar", None,
                                             -1))
        self.actionPan.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Help", None,
                                             -1))
        self.actionPan.setShortcut(
            QtWidgets.QApplication.translate("MainWindowPlot", "P", None, -1))
        self.actionZoom.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Zoom", None,
                                             -1))
        self.actionZoom.setToolTip(
            QtWidgets.QApplication.translate("MainWindowPlot", "Zoom control",
                                             None, -1))
        self.actionZoom.setShortcut(
            QtWidgets.QApplication.translate("MainWindowPlot", "Z", None, -1))
        self.actionConnectors.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Pins", None,
                                             -1))
        self.actionConnectors.setToolTip(
            QtWidgets.QApplication.translate(
                "MainWindowPlot",
                "Show connector pins for selected qcomponents", None, -1))
        self.actionConnectors.setShortcut(
            QtWidgets.QApplication.translate("MainWindowPlot", "C", None, -1))
        self.actionCoords.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Get point",
                                             None, -1))
        self.actionCoords.setToolTip(
            QtWidgets.QApplication.translate(
                "MainWindowPlot",
                "Click for position --- Enable this to click on the plot and log the (x,y) position",
                None, -1))
        self.actionCoords.setShortcut(
            QtWidgets.QApplication.translate("MainWindowPlot", "P", None, -1))
        self.actionAuto.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Autoscale",
                                             None, -1))
        self.actionAuto.setToolTip(
            QtWidgets.QApplication.translate("MainWindowPlot", "Auto Zoom",
                                             None, -1))
        self.actionAuto.setShortcut(
            QtWidgets.QApplication.translate("MainWindowPlot", "A", None, -1))
        self.actionReplot.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Replot", None,
                                             -1))
        self.actionReplot.setShortcut(
            QtWidgets.QApplication.translate("MainWindowPlot", "Ctrl+R", None,
                                             -1))
        self.actionRuler.setText(
            QtWidgets.QApplication.translate("MainWindowPlot", "Ruler", None,
                                             -1))
        self.actionRuler.setToolTip(
            QtWidgets.QApplication.translate("MainWindowPlot",
                                             "Activate the ruler", None, -1))


from .widgets.bases.expanding_toolbar import QToolBarExpanding
from . import main_window_rc_rc



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

#  This a launch structure used on BlueJayV2, used for wire bonding
#  There is no CPW tee attached to this p#

# Imports required for drawing

# import numpy as np # (currently not used, may be needed later for component customization)
from qiskit_metal import draw
from qiskit_metal.toolbox_python.attr_dict import Dict
from qiskit_metal.qlibrary.core import QComponent

# Define class and options for the launch geometry


class LaunchpadWirebondDriven(QComponent):
    r"""Launch pad to feed/read signals to/from the chip.

    Inherits 'QComponent' class.

    Creates a 50 ohm launch pad with a ground pocket cutout. Parameters control
    the launchpad polygons. The (0,0) point is the center of the necking of the
    launch tip. The pin attaches directly to the built-in lead length at its
    midpoint. There is also a pin at the back of the pad for DrivenModal simulations.

    Pocket and pad:
    - Pocket and launch pad geometries are currently fixed.
    - (0,0) point is the midpoint of the necking of the launch tip.
    - Pocket is a negative shape that is cut out of the ground plane.

    Values (unless noted) are strings with units included, (e.g., '30um').

    Sketch:
    Below is a sketch of the launch::

            -----------
            |          \
            |      ---------\\
          in|      |    0    |    (0,0) pin at midpoint of necking, before the lead
            |      ---------//
            |          /
            -----------

            y
            ^
            |
            |------> x

    .. image::
        LaunchpadWirebond.png

    .. meta::
        :description: Launchpad Wirebond Driven

    Default Options:
        * trace_width: 'cpw_width' -- Width of the transmission line attached to the launch pad
        * trace_gap: 'cpw_gap' -- Gap of the transmission line
        * lead_length: '25um' -- Length of the transmission line attached to the launch pad
        * pad_width: '80um' -- Width of the launch pad
        * pad_height: '80um' -- Height of the launch pad
        * pad_gap: '58um' -- Gap of the launch pad
        * taper_height: '122um' -- Height of the taper from the launch pad to the transmission line
    """

    default_options = Dict(trace_width='cpw_width',
                           trace_gap='cpw_gap',
                           lead_length='25um',
                           pad_width='80um',
                           pad_height='80um',
                           pad_gap='58um',
                           taper_height='122um')
    """Default options"""

    TOOLTIP = """Launch pad to feed/read signals to/from the chip."""

    def make(self):
        """This is executed by the user to generate the qgeometry for the
        component."""

        p = self.p

        pad_width = p.pad_width
        pad_height = p.pad_height
        pad_gap = p.pad_gap
        trace_width = p.trace_width
        trace_width_half = trace_width / 2.
        pad_width_half = pad_width / 2.
        lead_length = p.lead_length
        taper_height = p.taper_height
        trace_gap = p.trace_gap

        pad_gap = p.pad_gap
        #########################################################

        # Geometry of main launch structure
        # The shape is a polygon and we prepare this point as orientation is 0 degree
        launch_pad = draw.Polygon([
            (0, trace_width_half), (-taper_height, pad_width_half),
            (-(pad_height + taper_height), pad_width_half),
            (-(pad_height + taper_height), -pad_width_half),
            (-taper_height, -pad_width_half), (0, -trace_width_half),
            (lead_length, -trace_width_half), (lead_length, trace_width_half),
            (0, trace_width_half)
        ])

        # Geometry pocket (gap)
        # Same way applied for pocket
        pocket = draw.Polygon([(0, trace_width_half + trace_gap),
                               (-taper_height, pad_width_half + pad_gap),
                               (-(pad_height + taper_height + pad_gap),
                                pad_width_half + pad_gap),
                               (-(pad_height + taper_height + pad_gap),
                                -(pad_width_half + pad_gap)),
                               (-taper_height, -(pad_width_half + pad_gap)),
                               (0, -(trace_width_half + trace_gap)),
                               (lead_length, -(trace_width_half + trace_gap)),
                               (lead_length, trace_width_half + trace_gap),
                               (0, trace_width_half + trace_gap)])

        # These variables are used to graphically locate the pin locations
        main_pin_line = draw.LineString([(lead_length, trace_width_half),
                                         (lead_length, -trace_width_half)])
        driven_pin_line = draw.LineString([
            (-(pad_height + taper_height + pad_gap), pad_width_half),
            (-(pad_height + taper_height + pad_gap), -pad_width_half)
        ])

        # Create polygon object list
        polys1 = [main_pin_line, driven_pin_line, launch_pad, pocket]

        # Rotates and translates all the objects as requested. Uses package functions in
        # 'draw_utility' for easy rotation/translation
        polys1 = draw.rotate(polys1, p.orientation, origin=(0, 0))
        polys1 = draw.translate(polys1, xoff=p.pos_x, yoff=p.pos_y)
        [main_pin_line, driven_pin_line, launch_pad, pocket] = polys1

        # Adds the object to the qgeometry table
        self.add_qgeometry('poly', dict(launch_pad=launch_pad), layer=p.layer)

        # Subtracts out ground plane on the layer its on
        self.add_qgeometry('poly',
                           dict(pocket=pocket),
                           subtract=True,
                           layer=p.layer)

        # Generates the pins
        self.add_pin('tie', main_pin_line.coords, trace_width)
        self.add_pin('in', driven_pin_line.coords, pad_width, gap=pad_gap)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
""" Sweep a qcomponent option, and get results of analysis."""
# pylint: disable=too-many-lines
from typing import Tuple, Union
import pandas as pd

from qiskit_metal import Dict
# pylint: disable=unused-import
# QHFSSRenderer used to describe types in arguments.
from qiskit_metal.renderers.renderer_ansys.hfss_renderer import QHFSSRenderer


class Sweeping():
    """The methods allow users to sweep a variable in a components's options.
    Need access to renderers which are registered in QDesign."""

    def __init__(self, design: 'QDesign'):
        """Give QDesign to this class so Sweeping can access the registered
        QRenderers.

        Args:
            design (QDesign): Used to access the QRenderers.
        """
        self.design = design

    @classmethod
    def option_value(cls, a_dict: Dict, search: str) -> str:
        """Get value from dict based on key.  This method is used for unknown
        depth, dict search, within a dict.

        Args:
            a_dict (Dict): Dictionary to get values from
            search (str): String to search for

        Returns:
            str: Value from the dictionary of the searched term.
        """
        value = a_dict[search]
        return value

    def error_check_sweep_input(self, qcomp_name: str, option_name: str,
                                option_sweep: list) -> Tuple[list, Dict, int]:
        """ Implement error checking of data for sweeping.

        Args:
            qcomp_name (str): Component that contains the option to be swept.
            option_name (str): The option within qcomp_name to sweep.
            option_sweep (list): Each entry in the list is a value
                                for option_name.

        Returns:
            list: Traverse the option Dict.
            addict.addict.Dict: Value from the dictionary of the searched key.
            int: Observation of searching for data from arguments.

            * 0 Error not detected in the input-data.
            * 1 qcomp_name not registered in design.
            * 2 option_name is empty.
            * 3 option_name is not found as key in Dict.
            * 4 option_sweep is empty, need at least one entry.
        """
        option_path = None
        a_value = None

        if len(option_sweep) == 0:
            return option_path, a_value, 4

        if option_name:
            option_path = option_name.split('.')
        else:
            return option_path, a_value, 2

        if qcomp_name in self.design.components.keys():
            qcomp_options = self.design.components[qcomp_name].options
        else:
            return option_path, a_value, 1

        a_value = qcomp_options

        # All but the last item in list.
        for name in option_path[:-1]:
            if name in a_value:
                a_value = self.option_value(a_value, name)
            else:
                self.design.logger.warning(f'Key="{name}" is not in dict.')
                return option_path, a_value, 3

        return option_path, a_value, 0

    def prep_drivenmodal_setup(self, setup_args: Dict) -> int:
        """User can pass arguments for method drivenmodal setup.  If not passed,
        method will use the options in HFSS default_options. The name of setup
        will be "Sweep_dm_setup".  If a setup named "Sweep_dm_setup" exists
        in the project, it will be deleted, and a new setup will be added
        with the arguments from setup_args.

        Assume: Error checking has already occurred for existence of Ansys,
        project, and HFSS eigenmode design has been connected to pinfo.

        Args:
            setup_args (Dict):  Maximum  keys used in setup_args.

        **setup_args** dict contents:
            * freq_ghz (float, optional): Frequency in GHz.  Defaults to 5.
            * max_delta_s (float, optional): Absolute value of maximum
                    difference in scattering parameter S. Defaults to 0.1.
            * max_passes (int, optional): Maximum number of passes.
              Defaults to 10.
            * min_passes (int, optional): Minimum number of passes.
              Defaults to 1.
            * min_converged (int, optional): Minimum number of converged
              passes.  Defaults to 1.
            * pct_refinement (int, optional): Percent refinement.
              Defaults to 30.
            * basis_order (int, optional): Basis order. Defaults to 1.

        Returns:
            int: The return code of status.
                * 0 Setup of "Sweep_dm_setup" added to design with setup_args.
                * 1 Look at warning message to determine which argument was of
                  the wrong data type.
                * 2 Look at warning message, a key in setup_args that was
                  not expected.
        """
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-return-statements

        a_hfss = self.design.renderers.hfss
        setup_args.name = "Sweep_dm_setup"
        a_hfss.pinfo.design.delete_setup(setup_args.name)

        for key, value in setup_args.items():
            if key == "name":
                continue  #For this method, "Sweep_dm_setup" used.
            if key == "freq_ghz":
                if isinstance(value, float) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "float")
                return 1
            if key == "max_delta_s":
                if isinstance(value, float) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "float")
                return 1
            if key == 'max_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'min_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'min_converged':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'basis_order':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'pct_refinement':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1

            self.design.logger.warning(
                f'The key={key} is not expected.  Do you have a typo?  '
                f'The setup was not added to design. '
                f'Sweep will not be implemented.')
            return 2

        a_hfss.new_ansys_setup(**setup_args)
        a_hfss.activate_ansys_setup(setup_args.name)
        return 0

    def prep_eigenmode_setup(self, setup_args: Dict) -> int:
        """User can pass arguments for method eigenmode setup.  If not passed,
        method will use the options in HFSS default_options. The name of setup
        will be "Sweep_em_setup".  If a setup named "Sweep_em_setup" exists
        in the project, it will be deleted, and a new setup will be added
        with the arguments from setup_args.

        Assume: Error checking has already occurred for existence of Ansys,
        project, and HFSS eigenmode design has been connected to pinfo.

        Args:
            setup_args (Dict):  Maximum  keys used in setup_args.

        **setup_args** dict contents:
            * min_freq_ghz (int, optional): Minimum frequency in GHz.
              Defaults to 1.
            * n_modes (int, optional): Number of modes. Defaults to 1.
            * max_delta_f (float, optional): Maximum difference in
              freq between consecutive passes.
              Defaults to 0.5.
            * max_passes (int, optional): Maximum number of passes.
              Defaults to 10.
            * min_passes (int, optional): Minimum number of passes.
              Defaults to 1.
            * min_converged (int, optional): Minimum number of converged
              passes.  Defaults to 1.
            * pct_refinement (int, optional): Percent refinement.
              Defaults to 30.
            * basis_order (int, optional): Basis order. Defaults to -1.

        Returns:
            int: The return code of status.
                * 0 Setup of "Sweep_em_setup" added to design with setup_args.
                * 1 Look at warning message to determine which argument was of
                  the wrong data type.
                * 2 Look at warning message, a key in setup_args that was
                  not expected.
        """
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-return-statements

        a_hfss = self.design.renderers.hfss
        setup_args.name = "Sweep_em_setup"
        a_hfss.pinfo.design.delete_setup(setup_args.name)

        for key, value in setup_args.items():
            if key == "name":
                continue  #For this method, "Sweep_em_setup" used.
            if key == "min_freq_ghz":
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == "n_modes":
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == "max_delta_f":
                if isinstance(value, float) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "float")
                return 1
            if key == 'max_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'min_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'min_converged':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'basis_order':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'pct_refinement':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1

            self.design.logger.warning(
                f'The key={key} is not expected.  Do you have a typo?  '
                f'The setup was not added to design. '
                f'Sweep will not be implemented.')
            return 2

        a_hfss.new_ansys_setup(**setup_args)
        a_hfss.activate_ansys_setup(setup_args.name)
        return 0

    def warning_for_setup(self, setup_args: Dict, key: str, data_type: str):
        """Give a warning based on key/value of Dict.

        Args:
            setup_args (Dict): Holds the key/value of setup arguments
                        that are of interest.
            key (str): Name of setup argument.
            data_type (str): The data type the argument should be.
        """
        self.design.logger.warning(
            f'The value for {key} should be a {data_type}. '
            f'The present value is {setup_args[key]}.  '
            f'Sweep will not be implemented.')

    def prep_q3d_setup(self, setup_args: Dict) -> int:
        """User can pass arguments for method q3d setup.  If not passed,
        method will use the options in Q3D renderer default_options.
        The name of setup will be "Sweep_q3d_setup".  If a setup named
        "Sweep_q3d_setup" exists in the project, it will be deleted, and
        a new setup will be added with the arguments from setup_args.

        Assume: Error checking has already occurred for existence of Ansys,
        project, and Q3d Extractor design has been connected to pinfo.

        Args:
            setup_args (Dict): Maximum  keys used in setup_args.

        **setup_args** dict contents:
            * freq_ghz (float, optional): Frequency in GHz. Defaults to 5..
            * name (str, optional): Name of solution setup.
                        Defaults to "Setup".
            * max_passes (int, optional): Maximum number of passes.
                        Defaults to 15.
            * min_passes (int, optional): Minimum number of passes.
                        Defaults to 2.
            * percent_error (float, optional): Error tolerance as percentage.
                        Defaults to 0.5.
            * save_fields (bool, optional): Whether or not to save fields.
                        Defaults to False.
            * enabled (bool, optional): Whether or not setup is enabled.
                        Defaults to True.
            * min_converged_passes (int, optional): Minimum number of
                        converged passes. Defaults to 2.
            * percent_refinement (int, optional): Refinement as a percentage.
                        Defaults to 30.
            * auto_increase_solution_order (bool, optional): Whether or not
                        to increase solution order automatically.
                        Defaults to True.
            * solution_order (str, optional): Solution order.
                        Defaults to 'High'.
            * solver_type (str, optional): Solver type.
                        Defaults to 'Iterative'.

        Returns:
            int:  The return code of status.
                * 0 Setup of "Sweep_q3d_setup" added to design with setup_args.
                * 1 Look at warning message to determine which argument was
                        of the wrong data type.
                * 2 Look at warning message, a key in setup_args that was not
                        expected.
        """
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-return-statements

        a_q3d = self.design.renderers.q3d
        setup_args.name = "Sweep_q3d_setup"
        a_q3d.pinfo.design.delete_setup(setup_args.name)

        for key, value in setup_args.items():
            if key == "name":
                continue  #For this method, "Sweep_q3d_setup" used.
            if key == "freq_ghz":
                if isinstance(value, float) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "float")
                return 1
            if key == 'max_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'min_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'percent_error':
                if isinstance(value, float) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "float")
                return 1
            if key == 'save_fields':
                if isinstance(value, bool) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "bool")
                return 1
            if key == 'enabled':
                if isinstance(value, bool) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "bool")
                return 1
            if key == 'min_converged_passes':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'percent_refinement':
                if isinstance(value, int) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "int")
                return 1
            if key == 'auto_increase_solution_order':
                if isinstance(value, bool) or value is None:
                    continue
                self.warning_for_setup(setup_args, key, "bool")
                return 1
            if key == 'solution_order':
                if (isinstance(value, str) and
                        value in ["High", "Normal", "Higher", "Highest"
                                 ]) or value is None:
                    continue
                self.design.logger.warning(
                    f'The value for solution_order should be a str '
                    f'in ["High", "Normal", "Higher", "Highest"]. '
                    f'The present value is {value}.  '
                    f'Sweep will not be implemented.')
                return 1
            if key == 'solver_type':
                if isinstance(value, str) or value is None:
                    continue

                self.warning_for_setup(setup_args, key, "str")
                return 1

            self.design.logger.warning(
                f'The key={key} is not expected.  Do you have a typo?  '
                f'The setup was not added to design. '
                f'Sweep will not be implemented.')
            return 2

        a_q3d.new_ansys_setup(**setup_args)
        a_q3d.activate_ansys_setup(setup_args.name)
        return 0

    def sweep_one_option_get_eigenmode_solution_data(
            self,
            qcomp_name: str,
            option_name: str,
            option_sweep: list,
            qcomp_render: list,
            endcaps_render: list,
            ignored_jjs_render: list,
            box_plus_buffer_render: bool = True,
            setup_args: Dict = None,
            leave_last_design: bool = True,
            design_name: str = "Sweep_Eigenmode") -> Tuple[Dict, int]:
        """
        Ansys must be open with inserted project. A design, "HFSS Design"
        with eigenmode solution-type will be inserted by this method.

        Args:
            qcomp_name (str): A component that contains the option to be
                                swept. Assume qcomp_name is in qcomp_render.
            option_name (str): The option within qcomp_name to sweep.
                                Follow details from
                                renderer in QHFSSRenderer.render_design.
            option_sweep (list): Each entry in the list is a value for
                                option_name.
            qcomp_render(list): List of components to render to Ansys.
            endcaps_render (list): Identify which kind of pins.
                                    Follow details from renderer in
                                    QHFSSRenderer.render_design.
            ignored_jjs_render (list): List of tuples of jj's that shouldn't
                                    be rendered.  Follow details from
                                    renderer in QHFSSRenderer.render_design.
            box_plus_buffer_render (bool): Either calculate a bounding box
                                    based on the location of rendered
                                    geometries or use chip size from design
                                    class.  Follow details from renderer in
                                    QHFSSRenderer.render_design.
                                    Default is True.
            setup_args (Dict): Hold the arguments for  Hfss eigenmode setup()
                                    as  key/values to pass to Ansys.
                                    If None, default Setup will be used.
            leave_last_design (bool): In HFSS, after the last sweep,
                                    should the design be cleared?
                                    Default is True.
            design_name (str, optional):  Name of HFSS_design to use in
                                    project. Defaults to "Sweep_Eigenmode".

        Returns:
            Tuple[Dict, int]: The dict key is each value of option_sweep, the
            value is the solution-data for each sweep.
            The int is the observation of searching for data from arguments as
            defined below.

            * 0 Have list of capacitance matrix.
            * 1 qcomp_name not registered in design.
            * 2 option_name is empty.
            * 3 option_name is not found as key in dict.
            * 4 option_sweep is empty, need at least one entry.
            * 5 last key in option_name is not in dict.
            * 6 project not in app
            * 7 design not in app
            * 8 setup not implement, check the setup_args.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-arguments

        #Dict of all swept information.
        all_sweep = Dict()
        option_path, a_value, check_result, = self.error_check_sweep_input(
            qcomp_name, option_name, option_sweep)
        if check_result != 0:
            return all_sweep, check_result

        a_hfss = self.design.renderers.hfss
        # Assume Ansys is open, with a project open.
        a_hfss.connect_ansys()
        a_hfss.activate_ansys_design(design_name, 'eigenmode')

        a_hfss.clean_active_design()

        if self.prep_eigenmode_setup(setup_args) != 0:
            self.design.logger.warning(
                'The setup was not implemented, look at warning messages.')
            return all_sweep, 8

        len_sweep = len(option_sweep) - 1

        for index, item in enumerate(option_sweep):
            if option_path[-1] in a_value.keys():
                a_value[option_path[-1]] = item
            else:
                self.design.logger.warning(
                    f'Key="{option_path[-1]}" is not in dict.')
                return all_sweep, 5

            self.design.rebuild()

            a_hfss.render_design(selection=qcomp_render,
                                 open_pins=endcaps_render,
                                 ignored_jjs=ignored_jjs_render,
                                 box_plus_buffer=box_plus_buffer_render
                                )  #Render the items chosen

            a_hfss.analyze_setup(
                a_hfss.pinfo.setup.name)  #Analyze said solution setup.
            setup = a_hfss.pinfo.setup
            #solution_name = setup.solution_name
            all_solutions = setup.get_solutions()
            #setup_names = all_solutions.list_variations()
            freqs, kappa_over_2pis = all_solutions.eigenmodes()
            df_t, df_f, _, convergence = self.hfss_em_get_convergence(a_hfss)

            if not convergence:
                self.design.logger.warning(
                    f'Heads-Up: {option_name}={item} Failed to converge.  ')

            sweep_values = Dict()

            sweep_values['convergence'] = convergence
            sweep_values['option_name'] = option_path[-1]
            sweep_values['frequency'] = freqs
            sweep_values['kappa_over_2pis'] = kappa_over_2pis
            sweep_values['quality_factor'] = self.get_quality_factor(
                freqs, kappa_over_2pis)

            sweep_values['convergence_eig_f'] = df_f
            sweep_values['convergence_t'] = df_t
            all_sweep[item] = sweep_values

            #Decide if need to clean the design.
            obj_names = a_hfss.pinfo.get_all_object_names()
            if obj_names:
                if index == len_sweep and not leave_last_design:
                    a_hfss.clean_active_design()
                elif index != len_sweep:
                    a_hfss.clean_active_design()

        a_hfss.disconnect_ansys()
        return all_sweep, 0

    def hfss_em_get_convergence(
        self,
        a_hfss: 'QHFSSRenderer',
        variation: str = None
    ) -> Tuple[pd.core.frame.DataFrame, pd.core.frame.DataFrame, str, bool]:
        """Use QHFSSRenderer to get convergence data from Ansys for eigenmode.

        Args:
            a_hfss (QHFSSRenderer): Reference to renderer to Ansys.
            variation (str, optional): Information from pyEPR; variation should
                        be in the form variation = "scale_factor='1.2001'".
                        Defaults to None.

        Returns:
            Tuple[pd.core.frame.DataFrame, pd.core.frame.DataFramestr, str, str]:
            1st DataFrame: Convergence_t
            2nd DataFrame: Convergence_f
            3rd list: Text from GUI of solution data.
            4th bool: If data converged.
        """
        convergence_t, convergence_f, gui_text = a_hfss.get_convergences(
            variation)

        text_list, convergence = self.parse_text_from_hfss_convergence(gui_text)

        return convergence_t, convergence_f, text_list, convergence

    def parse_text_from_hfss_convergence(self,
                                         gui_text: str) -> Tuple[list, bool]:
        """Parse the text obtained from hfss after analysis.

        Args:
            gui_text (str): Text from GUI of solution data.

        Returns:
            Tuple[list, bool]: All parsed information
            1st list: Text from GUI of solution data.
            2nd bool: If data converged.
        """
        text_list = gui_text.splitlines()

        # Find convergence information in text.
        convergence_bool = [
            string for string in text_list if 'Converged' in string
        ]
        if len(convergence_bool) == 1:
            if 'Yes' in convergence_bool[0]:
                convergence = True
            else:
                convergence = False
        else:
            convergence = False
            self.design.logger.warning(
                'Either could not find Converged information or too many '
                'entries in text. Force convergence to be False.')
        return text_list, convergence

    def hfss_dm_get_convergence(
            self,
            a_hfss: 'QHFSSRenderer',
            variation: str = None) -> Tuple[pd.core.frame.DataFrame, str, bool]:
        """Use QHFSSRenderer to get convergence data from Ansys for drivenmodal.

        Args:
            a_hfss (QHFSSRenderer): Reference to renderer to Ansys.
            variation (str, optional): Information from pyEPR; variation should
                            be in the form variation = "scale_factor='1.2001'".
                            Defaults to None.

        Returns:
            Tuple[pd.core.frame.DataFrame, str, bool]:
            1st DataFrame: Convergence_t
            2nd list: Text from GUI of solution data.
            3rd bool: If data converged.
        """

        convergence_t, _, gui_text = a_hfss.get_convergences(variation)
        text_list, convergence = self.parse_text_from_hfss_convergence(gui_text)
        return convergence_t, text_list, convergence

    def sweep_one_option_get_drivenmodal_solution_data(
            self,
            qcomp_name: str,
            option_name: str,
            option_sweep: list,
            dm_render_args: Dict,
            dm_add_sweep_args: Dict,
            setup_args: Dict = None,
            leave_last_design: bool = True,
            design_name: str = "Sweep_DrivenModal") -> Tuple[Dict, int]:
        """
        Ansys must be open with inserted project. A design, "HFSS Design"
        with Driven Modal solution-type will be inserted by this method.

        Args:
            qcomp_name (str): A component that contains the option to be
                                swept. Assume qcomp_name is in qcomp_render.
            option_name (str): The option within qcomp_name to sweep.
            option_sweep (list): Each entry in the list is a value for
                                option_name.
            dm_render_args (Dict): Arguments to pass to render_design().
                                Next six items are key/value pairs.
            *selection(list): List of components to render to Ansys.
            *open_pins (list): Identify which kind of pins.
                                    Follow details from renderer in
                                    QHFSSRenderer.render_design.
            *port_list (list): List of tuples of jj's that shouldn't
                                    be rendered.  Follow details from
                                    renderer in QHFSSRenderer.render_design.
            *jj_to_port (list): List of junctions (qcomp, qgeometry_name,
                                impedance, draw_ind) to render as lumped ports
                                or as lumped port in parallel with a sheet
                                inductance.    Follow details from renderer
                                in QHFSSRenderer.render_design.
            *ignored_jjs (list): List of junctions (qcomp, qgeometry_name) to
                                omit altogether during rendering.   Follow
                                details from renderer in
                                QHFSSRenderer.render_design.
            *box_plus_buffer (bool): Either calculate a bounding box
                                    based on the location of rendered
                                    geometries or use chip size from design
                                    class.  Follow details from renderer in
                                    QHFSSRenderer.render_design.
                                    Default is True.
            dm_add_sweep_args(Dict): Arguments to pass to insert_sweep() in
                                    pyEPR, through add_sweep in QHFSSRenderer.
                                    Next 7 items are key/value pairs, if passed
                                    will be used. start_ghz and
                                    stop_ghz must be passed. Follow details from
                                    insert_sweep()  in pyEPR.  If an allowable key is not
                                    passed, the default will be used.  You can
                                    provide either step_ghz OR count when
                                    inserting an HFSS driven model freq sweep.
                                    DO NOT provide both OR neither!

            *start_ghz (float): Starting frequency of sweep in GHz.
            *stop_ghz (float):  Ending frequency of sweep in GHz.
            *count(int):  Total number of frequencies.  Defaults to 101.
            step_ghz ([type]): Difference between adjacent frequencies.
                            Defaults to None.
            *name (str): Name of sweep. Defaults to "Sweep_options__dm_sweep".
            *type(str): Type of sweep.   Defaults to "Fast". Choose from
                        'Fast', 'Interpolating', 'Discrete'.
            *save_fields (bool): Whether or not to save fields.
                                 Defaults to False.

            setup_args (Dict): Hold the arguments for  Hfss driven-modal setup()
                                    as  key/values to pass to Ansys.
                                    If None, default Setup will be used.

            leave_last_design (bool): In HFSS, after the last sweep,
                                    should the design be cleared?
                                    Default is True.
            design_name (str, optional):  Name of HFSS_design to use in
                                    project. Defaults to "Sweep_DrivenModal".

        Returns:
            Tuple[Dict, int]: The dict key is each value of option_sweep, the
            value is the solution-data for each sweep.
            There is a pandas dataframe for Scatter matrix, Impedance matrix,
            and Admittance matrix.

            The int is the observation of searching for data from arguments as
            defined below.

            * 0 Have list of capacitance matrix.
            * 1 qcomp_name not registered in design.
            * 2 option_name is empty.
            * 3 option_name is not found as key in Dict.
            * 4 option_sweep is empty, need at least one entry.
            * 5 last key in option_name is not in Dict.
            * 6 project not in app
            * 7 design not in app
            * 8 setup not implement, check the setup_args.
            * 9 dm_render_args is missing keys in the Dict.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-arguments

        #Dict of all swept information.
        all_sweep = Dict()
        option_path, a_value, check_result, = self.error_check_sweep_input(
            qcomp_name, option_name, option_sweep)
        if check_result != 0:
            return all_sweep, check_result

        a_hfss = self.design.renderers.hfss
        # Assume Ansys is open, with a project open.
        a_hfss.connect_ansys()
        a_hfss.activate_ansys_design(design_name, 'drivenmodal')

        a_hfss.clean_active_design()

        if self.prep_drivenmodal_setup(setup_args) != 0:
            self.design.logger.warning(
                'The setup was not implemented, look at warning messages.')
            return all_sweep, 8

        if self.error_check_render_design_args(dm_render_args) != 0:
            return all_sweep, 9

        len_sweep = len(option_sweep) - 1

        for index, item in enumerate(option_sweep):
            if option_path[-1] in a_value.keys():
                a_value[option_path[-1]] = item
            else:
                self.design.logger.warning(
                    f'Key="{option_path[-1]}" is not in dict.')
                return all_sweep, 5

            self.design.rebuild()

            a_hfss.render_design(selection=dm_render_args.selection,
                                 open_pins=dm_render_args.open_pins,
                                 port_list=dm_render_args.port_list,
                                 jj_to_port=dm_render_args.jj_to_port,
                                 ignored_jjs=dm_render_args.ignored_jjs,
                                 box_plus_buffer=dm_render_args.box_plus_buffer)

            #To insert a "frequency sweep" within setup,
            # the pin/ports have to be rendered.
            self.error_check_and_insert_sweep(a_hfss, setup_args.name,
                                              dm_add_sweep_args)

            matrix_size = self.get_size_of_matrix(dm_render_args)

            self.populate_dm_all_sweep(all_sweep, a_hfss, dm_add_sweep_args,
                                       setup_args, matrix_size, item,
                                       option_name)
            #Decide if need to clean the design.
            obj_names = a_hfss.pinfo.get_all_object_names()
            if obj_names:
                if index == len_sweep and not leave_last_design:
                    a_hfss.clean_active_design()
                elif index != len_sweep:
                    a_hfss.clean_active_design()

        a_hfss.disconnect_ansys()
        return all_sweep, 0

    def populate_dm_all_sweep(self, all_sweep: Dict, a_hfss: 'QHFSSRenderer',
                              dm_add_sweep_args: Dict, setup_args: Dict,
                              matrix_size: int, item: str, option_name: str):
        """The Dict all_sweep holds three matrices for each iteration of a
        option in Metal.

        Args:
            all_sweep (Dict): To hold the output for each iteration of option
                             from Metal.
            a_hfss (QHFSSRenderer): Reference to Metal Ansys renderer.
            dm_add_sweep_args (Dict): Has name of sweep in setup.
            setup_args (Dict): Has name of setup.
            matrix_size (int): Size of matrix to retreive from Ansys sweep.
            item (str): Each value of option in Metal to iterate through.
            option_name (str): The option within qcomp_name to sweep.
        """
        sweep_values = Dict()
        if matrix_size == 0:
            sweep_values['s_matrix'] = None
            sweep_values['y_matrix'] = None
            sweep_values['z_matrix'] = None
        else:
            a_hfss.analyze_sweep(dm_add_sweep_args.name, setup_args.name)
            # pylint: disable=invalid-name
            s_Pparms, y_Pparams, z_Pparams = a_hfss.get_all_Pparms_matrices(
                matrix_size)

            convergence_t, dummy_text_list, convergence = self.hfss_dm_get_convergence(
                a_hfss, variation=None)
            sweep_values['option_name'] = option_name
            sweep_values['convergence'] = convergence
            sweep_values['s_matrix'] = s_Pparms
            sweep_values['y_matrix'] = y_Pparams
            sweep_values['z_matrix'] = z_Pparams
            sweep_values['convergence_t'] = convergence_t
        all_sweep[item] = sweep_values

    @classmethod
    def get_size_of_matrix(cls, dm_render_args: Dict) -> int:
        """Determine the size of s_matrix, y_matrix, z_matrix.
        s_matrix =
        size of list of pins to render to lumped port +
        size of list of junctions to render as lumped port.

        size of matrix = size of 3rd parameter + size of fourth parfameter

        List of arguments for render_design:
            - First parameter: List of components to render
                            (empty list if rendering whole Metal design)
            - Second parameter: List of pins (qcomp, pin) with open endcaps
            - Third parameter: List of pins (qcomp, pin, impedance) to render as lumped ports
            - Fourth parameter: List of junctions (qcomp, qgeometry_name, impedance, draw_ind)
               to render as lumped ports or as lumped port in parallel with a sheet inductance
            - Fifth parameter: List of junctions (qcomp, qgeometry_name) to
                            omit altogether during rendering
            - Sixth parameter: Whether to render chip via box plus buffer or fixed chip size

        Args:
            dm_render_args (Dict): Holds the arguments used for render_design.

        Returns:
            int: Size of expected S-matrix.
        """
        matrix_size = 0
        if dm_render_args.port_list:
            matrix_size += len(dm_render_args.port_list)
        if dm_render_args.jj_to_port:
            matrix_size += len(dm_render_args.jj_to_port)

        return matrix_size

    def error_check_and_insert_sweep(
            self,
            a_hfss: 'QHFSSRenderer',
            setup_name: str,
            dm_add_sweep_args: Dict,
            dm_sweep_name: str = 'Sweep_options__dm_sweep'):
        """To insert an Ansys sweep to the named setup.  The Dict
        dm_add_sweep_args comes from user. They are arguments to pass to
        add sweep to Ansys.  If there are unexpected arguments, they will
        be removed.

        Args:
            a_hfss (QHFSSRenderer): Reference to Ansys renderer.
            setup_name (str): Name of active setup in active project.
            dm_add_sweep_args (Dict): Arguments to pass to
                        QHFSSRenderer.add_sweep.  If key is not in Dict, the
                        default value will be used.  The keys have to match
                        arguments which are passed to insert_sweep() in pyEPR,
                        through add_sweep in QHFSSRenderer.
            dm_sweep_name (str, optional): Name of inserted Ansys sweep.
                                    qDefaults to 'Sweep_options__dm_sweep'.
        """
        if 'name' not in dm_add_sweep_args.keys():
            dm_add_sweep_args['name'] = dm_sweep_name

        allowed_keys = {
            'start_ghz', 'stop_ghz', 'count', 'step_ghz', 'name', 'type',
            'save_fields'
        }
        unexpected_keys = set(dm_add_sweep_args.keys()) - allowed_keys

        if unexpected_keys:
            #[dm_add_sweep_args.pop(key) for key in unexpected_keys]
            self.design.logger.warning(
                f'Removed keys: {unexpected_keys} from '
                'dm_add_sweep_args Dict before using it. ')

        # pylint: disable=protected-access
        all_sweep_names = a_hfss.pinfo.setup._setup_module.GetSweeps(setup_name)
        if dm_add_sweep_args.name not in all_sweep_names:
            a_hfss.add_sweep(setup_name=setup_name, **dm_add_sweep_args)

    def error_check_render_design_args(self, dm_render_args: Dict) -> int:  # pylint: disable=too-many-return-statements
        """To render to Ansys, we need every argument in render_design.
        This method confirms all arguments are present.

        Args:
            dm_render_args (Dict): Dict that has the key=argument name,
                        value=data for argument.

        Returns:
            int: Return code.
            * 0 All the expected keys in Dict.
            * 1 A key is missing in dm_render_args Dict, look at warning message.
        """
        all_keys = dm_render_args.keys()

        if 'selection' not in all_keys:
            self.design.logger.warning(
                'The key selection is missing in Dict dm_render_args. '
                'Method render_design() NOT implemented.')
            return 1
        if 'open_pins' not in all_keys:
            self.design.logger.warning(
                'The key open_pins is missing in Dict dm_render_args. '
                'Method render_design() NOT implemented.')
            return 1
        if 'port_list' not in all_keys:
            self.design.logger.warning(
                'The key port_list is missing in Dict dm_render_args. '
                'Method render_design() NOT implemented.')
            return 1
        if 'jj_to_port' not in all_keys:
            self.design.logger.warning(
                'The key jj_to_port is missing in Dict dm_render_args. '
                'Method render_design() NOT implemented.')
            return 1
        if 'ignored_jjs' not in all_keys:
            self.design.logger.warning(
                'The key ignored_jjs is missing in Dict dm_render_args. '
                'Method render_design() NOT implemented.')
            return 1
        if 'box_plus_buffer' not in all_keys:
            self.design.logger.warning(
                'The key box_plus_buffer is missing in Dict dm_render_args. '
                'Method render_design() NOT implemented.')
            return 1
        return 0

    def get_quality_factor(
            self,
            freqs: Union[list, None] = None,
            kappa_over_2pis: Union[list, None] = None) -> Union[list, None]:
        """Calculate Quality Factor = freqs/kappa_over_2pis.  Before division,
        some error checking.

        Args:
            freqs (Union[list, None], optional): The eigenmode frequency.
                                                Defaults to None.
            kappa_over_2pis (Union[list, None], optional): The kappa/(2*pi).
                                                Defaults to None.

        Returns:
            Union[list, None]: Calculate freqs/kappa_over_2pis
        """

        quality_factor = None
        if kappa_over_2pis is None:
            return quality_factor

        # Assume both are lists or None since method:  eigenmodes()
        #                              in pyEPR returns a list or None.
        if len(freqs) == len(kappa_over_2pis):
            quality_factor = [
                float(ff) / float(kk) for ff, kk in zip(freqs, kappa_over_2pis)
            ]
            return quality_factor

        self.design.logger.warning(
            'The Quality factor not calculated since size of freqs'
            ' and kappa_over_2pis are not identical')
        return quality_factor

    def sweep_one_option_get_capacitance_matrix(
            self,
            qcomp_name: str,
            option_name: str,
            option_sweep: list,
            qcomp_render: list,
            endcaps_render: list,
            setup_args: Dict = None,
            leave_last_design: bool = True,
            design_name: str = "Sweep_Capacitance") -> Tuple[Dict, int]:
        """Ansys must be open with an inserted project.  A design,
        "Q3D Extractor Design", will be inserted by this method.

        Args:
            qcomp_name (str): A component that contains the option to be swept.
            option_name (str): The option within qcomp_name to sweep.
            option_sweep (list): Each entry in the list is a value for
                        option_name.
            qcomp_render (list): The component to render to Q3D.
            endcaps_render (list): Identify which kind of pins. Follow the
                        details from renderer QQ3DRenderer.render_design.
            setup_args (Dict): Hold the arguments for  Q3d setup() as
                        key/values to pass to Ansys.
                        If None, default Setup will be used.
            leave_last_design (bool) : In Q3d, after the last sweep, should
                        the design be cleared?
            design_name(str): Name of q3d_design to use in project.

        Returns:
            Dict or int: If dict, the key is each value of option_sweep, the
            value is the capacitance matrix for each sweep.
            If int, observation of searching for data from arguments as
            defined below:

            * 0 Have list of capacitance matrix.
            * 1 qcomp_name not registered in design.
            * 2 option_name is empty.
            * 3 option_name is not found as key in Dict.
            * 4 option_sweep is empty, need at least one entry.
            * 5 last key in option_name is not in Dict.
            * 6 project not in app
            * 7 design not in app
            * 8 setup not implement, check the setup_args.

        """
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-locals

        #Dict of all swept information.
        all_sweep = Dict()
        option_path, a_value, check_result = self.error_check_sweep_input(
            qcomp_name, option_name, option_sweep)
        if check_result != 0:
            return all_sweep, check_result

        a_q3d = self.design.renderers.q3d
        # Assume Ansys is open, with a project open.
        a_q3d.connect_ansys()
        a_q3d.activate_ansys_design(design_name, 'capacitive')

        a_q3d.clean_active_design()

        # Add a solution setup.
        if self.prep_q3d_setup(setup_args) != 0:
            self.design.logger.warning('The setup was not implemented, '
                                       'please look at warning messages.')
            return all_sweep, 8

        len_sweep = len(option_sweep) - 1

        # Last item in list.
        for index, item in enumerate(option_sweep):
            if option_path[-1] in a_value.keys():
                a_value[option_path[-1]] = item
            else:
                self.design.logger.warning(
                    f'Key="{option_path[-1]}" is not in dict.')
                return all_sweep, 5

            self.design.rebuild()

            a_q3d.render_design(
                selection=qcomp_render,
                open_pins=endcaps_render)  #Render the items chosen

            self.populate_q3d_all_sweep(all_sweep, a_q3d, item, option_name)

            #Decide if need to clean the design.
            obj_names = a_q3d.pinfo.get_all_object_names()
            if obj_names:
                if index == len_sweep and not leave_last_design:
                    a_q3d.clean_active_design()
                elif index != len_sweep:
                    a_q3d.clean_active_design()

        a_q3d.disconnect_ansys()
        return all_sweep, 0

    def populate_q3d_all_sweep(self, all_sweep: Dict, a_q3d: 'QQ3DRenderer',
                               item: str, option_name: str):
        """Analyze the setup, get convergence and capacitance information,
        populate all_sweep for each item which corresponds to option_name.

        Args:
            all_sweep (Dict): Reference to hold each item which corresponds
              to option_name.
            a_q3d (QQ3DRenderer): Reference to Ansys-Q3D renderer.
            item (str): The value of each item that we want to sweep
              in option_name.
            option_name (str): The option of QComponent that we want to sweep.
        """

        #Analyze said solution setup.
        a_q3d.analyze_setup(a_q3d.pinfo.setup.name)

        # If 'LastAdaptive' is used, then the pass_number won't affect anything.
        # If 'AdaptivePass' is used, then the pass_number is used.
        convergence_df, convergence_txt = a_q3d.pinfo.setup.get_convergence()
        target, current, pass_min = self._parse_text_from_q3d_convergence(
            convergence_txt)
        is_converged = self._test_if_q3d_analysis_converged(
            target, current, pass_min)
        cap_matrix = a_q3d.get_capacitance_matrix(variation='',
                                                  solution_kind='LastAdaptive',
                                                  pass_number=1)

        sweep_values = Dict()
        sweep_values['option_name'] = option_name
        #sweep_values['option_name'] = option_path[-1]
        sweep_values['convergence_target'] = target
        sweep_values['convergence_current'] = current
        sweep_values['min_number_passes'] = pass_min
        sweep_values['is_convergence'] = is_converged
        sweep_values['capacitance'] = cap_matrix
        sweep_values['convergence_data'] = convergence_df
        all_sweep[item] = sweep_values

    @classmethod
    def _test_if_q3d_analysis_converged(cls, target: float, current: float,
                                        passes_min: int) -> Union[bool, None]:
        """Use solution-data from Ansys-Q3d to determine if converged.

        Args:
            target (float): Delta percentage for target. Default is None.
            current (float): Delta percentage for current. Default is None.
            passes_min (int): Regarding convergence, minimum number of passes.
              Default is None.

        Returns:
            Union[bool, None]: Had solution data converged.  Default is None.
        """

        if None not in (target, current, passes_min):
            # Confirm that all three numbers have an value.
            if current <= target and passes_min > 1:
                is_converged = True
                return is_converged
            is_converged = False
            return is_converged

        is_converged = None
        return is_converged

    def _parse_text_from_q3d_convergence(
            self,
            gui_text: str) -> Tuple[Union[None, float], Union[None, float]]:
        """Parse gui_text using a priori known formatting. Ansys-Q3D
        solution-data provides gui_text.

        Args:
            gui_text (str): From Ansys-GUI-SolutionData.

        Returns:
            1st Union[None, float]: Delta percentage for target. Default is None.
            2nd Union[None, float]: Delta percentage for current. Default is None.
        """

        text_list = gui_text.splitlines()

        # Find Target information in text.
        target_all = [string for string in text_list if 'Target' in string]

        # Find Current information in text.
        current_all = [string for string in text_list if 'Current' in string]

        # Find Minimum number of passes from solution-data.
        min_passes_all = [string for string in text_list if 'Minimum' in string]

        target = self._extract_target_delta(target_all)
        current = self._extract_current_delta(current_all)
        min_passes = self._extract_min_passes(min_passes_all)

        return target, current, min_passes

    def _extract_min_passes(self, min_passes_all: list) -> Union[None, float]:
        """Given a pre-formatted list, search and return the "Minimum Number
        Of Passes."

        Args:
            min_passes_all (list): Result of search through string returned from Ansys-Q3D.

        Returns:
            Union[None, float]: Regarding convergence, minimum number of passes.
              Default is None.
        """

        min_num_of_passes = None
        if len(min_passes_all) == 1:
            if min_passes_all[0]:
                _, _, min_passes_str = min_passes_all[0].partition(':')
                try:
                    min_num_of_passes = int(min_passes_str)
                except ValueError:
                    self.design.logger.warning(
                        f'Target={min_passes_str} in GUI is not an int.'
                        'Force Minimum Number Of Passes to be None.')
        else:
            self.design.logger.warning(
                'Either could not find Minimum Number of Passes '
                'information or too many entries in text. '
                'Force Minimum Number of Passes to be None.')
        return min_num_of_passes

    def _extract_target_delta(self, target_all: list) -> Union[None, float]:
        """Given a pre-formatted list, search and return the target-delta
        percentage for convergence.

        Args:
            target_all (list): Result of search through string returned from Ansys-Q3D.

        Returns:
            Union[None, float]: Delta percentage for target. Default is None.
        """

        target = None
        if len(target_all) == 1:
            if target_all[0]:
                _, _, target_str = target_all[0].partition(':')
                try:
                    target = float(target_str)
                except ValueError:
                    self.design.logger.warning(
                        f'Target={target_str} in GUI is not a float.'
                        'Force Target Delta to be None.')
        else:
            self.design.logger.warning(
                'Either could not find Target Delta information or too many '
                'entries in text. Force Target Delta to be None.')
        return target

    def _extract_current_delta(self, current_all: list) -> Union[None, float]:
        """Given a pre-formatted list, search and return the current-delta
        percentage for convergence.

        Args:
            current_all (list): Result of search through string returned from Ansys-Q3D.

        Returns:
            Union[None, float]: Delta percentage for current. Default is None.
        """

        current = None
        if len(current_all) == 1:
            if current_all[0]:
                _, _, current_str = current_all[0].partition(':')
                try:
                    current = float(current_str)
                except ValueError:
                    self.design.logger.warning(
                        f'Target={current_str} in GUI is not a float.'
                        'Force Current Delta to be None.')
        else:
            self.design.logger.warning(
                'Either could not find Current Delta information or too many '
                'entries in text. Force Current Delta to be None.')
        return current


# The methods allow users to sweep a variable in a components's options.



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""File contains dictionary for NSquareSpiral and the make()."""

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent
import numpy as np


class ResonatorCoilRect(QComponent):
    """A rectangle spiral resonator based on length input. The X dimension is
    modified by the code based on the total length inputed.

    Inherits `QComponent` class

    A rectangular spiral resonator. The width of the spiral is modified based on inputted values
    and given total length of the spiral.

        ::

            <--------X-------->
            __________________
            |   ___________   |
            |  |           |  |
            |  |           |  |
            |  |______________|
            |

    .. image::
        ResonatorCoilRect.png

    .. meta::
        :description: Resonator Coil Rectangle

    Default Options:
        Convention: Values (unless noted) are strings with units included,
        (e.g., '30um')

        * n: '3' -- Number of turns of the spiral
        * length: '2000um' -- Total length of the spiral
        * line_width: '1um' -- The width of the line of the spiral
        * height: '40um' -- The height of the inner portion of the spiral
        * gap: '4um' -- The distance between each layer of the spiral
        * coupler_distance: '10um' -- The pin position from the grounded termination of the spiral
    """
    component_metadata = Dict(short_name='res')
    """Component metadata"""

    default_options = Dict(n='3',
                           length='2000um',
                           line_width='1um',
                           height='40um',
                           gap='4um',
                           coupler_distance='10um')
    """Default drawing options"""

    TOOLTIP = """A rectangle spiral resonator based on length input. The X dimension is
    modified by the code based on the total length inputed."""

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc."""
        p = self.p  # p for parsed parameters. Access to the parsed options.
        n = int(p.n)
        # Create the geometry

        spiral_list = []

        #Formulat to determine the size of the spiral based on inputed length.
        x_n = (p.length / (2 * n)) - (p.height + 2 * (p.gap + p.line_width) *
                                      (2 * n - 1))

        if x_n <= p.gap + p.line_width:
            self._error_message = 'Inputted values results in the width of the spiral being too small.'
            self.logger.warning(self._error_message)
            return

        for step in range(n):
            x_point = x_n / 2 + step * (p.line_width + p.gap)
            y_point = p.height / 2 + step * (p.line_width + p.gap)

            spiral_list.append((-x_point, -y_point))
            spiral_list.append((x_point, -y_point))
            spiral_list.append((x_point, y_point))
            spiral_list.append((-x_point - (p.line_width + p.gap), y_point))

        x_point = (x_n / 2 + (step + 1) * (p.line_width + p.gap))
        y_point = (p.height / 2 + (step + 1) * (p.line_width + p.gap) -
                   p.line_width / 2)
        spiral_list.append((-x_point, -y_point))
        spiral_list = draw.LineString(spiral_list)

        spiral_etch = draw.shapely.geometry.box(
            -(x_point + p.line_width / 2 + p.gap), -y_point,
            x_point - p.line_width / 2, y_point)
        #Generates a linestring to track port location
        points = draw.LineString([
            (-x_point + p.line_width / 2, -y_point + p.coupler_distance),
            (-x_point - p.line_width / 2, -y_point + p.coupler_distance)
        ])

        c_items = [spiral_list, spiral_etch, points]
        c_items = draw.rotate(c_items, p.orientation, origin=(0, 0))
        c_items = draw.translate(c_items, p.pos_x, p.pos_y)
        [spiral_list, spiral_etch, points] = c_items
        ##############################################
        # add elements
        self.add_qgeometry('path', {'n_spiral': spiral_list},
                           width=p.line_width)
        self.add_qgeometry('poly', {'n_spira_etch': spiral_etch}, subtract=True)

        # NEW PIN SPOT
        self.add_pin('spiralPin',
                     points=np.array(points.coords),
                     width=p.line_width,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""MPL Renderer."""
import logging
import random
import sys
from typing import TYPE_CHECKING, List

import matplotlib as mpl
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cycler import cycler
from .patch import PolygonPatch
from IPython.display import display
from matplotlib.axes import Axes
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from matplotlib.cbook import _OrderedSet
from matplotlib.collections import LineCollection, PatchCollection
from matplotlib.figure import Figure
from matplotlib.transforms import Bbox

from shapely.geometry import CAP_STYLE, JOIN_STYLE, LineString

from ... import Dict
from ...designs import QDesign
from .mpl_interaction import MplInteraction, PanAndZoom
from .mpl_toolbox import _axis_set_watermark_img, clear_axis, get_prop_cycle

from .. import config
if not config.is_building_docs():
    from ...toolbox_python.utility_functions import log_error_easy
    from qiskit_metal.toolbox_python.utility_functions import bad_fillet_idxs

if TYPE_CHECKING:
    from ..._gui.main_window import MetalGUI
    from ..._gui.widgets.plot_widget.plot_window import QMainWindowPlot
    from .mpl_canvas import PlotCanvas
    from qiskit_metal.elements.elements_handler import QGeometryTables

__all__ = ['QMplRenderer']

to_poly_patch = np.vectorize(PolygonPatch)


class QMplRenderer():
    """Matplotlib renderer for Metal designs.

    Notes:
        Access via ``gui.canvas.metal_renderer``. The axis to render is passed
        in the ``render`` method.
    """

    def __init__(self, canvas: 'PlotCanvas', design: QDesign,
                 logger: logging.Logger):
        """
        Args:
            canvas (PlotCanvas): The canvas
            design (QDesign): The design
            logger (logging.Logger): The logger
        """
        super().__init__()
        self.logger = logger
        self.canvas = canvas
        self.ax = None
        self.design = design
        self.options = Dict(resolution='16',)

        # Filter view options
        self.hidden_layers = set()

        # Set of component ids which are integers.
        self._hidden_components = set()

        self.colors = [
            '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b',
            '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'
        ]

        self.set_design(design)

    def get_color_num(self, num: int) -> str:
        """Get the color from the given number.
        Args:
            num (int): number
        Return:
            str: color
        """
        return self.colors[num % len(self.colors)]

    def hide_component(self, name):
        """Hide the component with the given name.
        Args:
            name (str): Component name
        """
        comp_id = self.design.components[name].id
        self._hidden_components.add(comp_id)

    def show_component(self, name):
        """Show the component with the given name.
        Args:
            name (str): Component name
        """
        comp_id = self.design.components[name].id
        self._hidden_components.discard(name)

    def hide_layer(self, name):
        """Hide the layer with the given name.
        Args:
            name (str): Layer name
        """
        self.hidden_layers.add(name)

    def show_layer(self, name):
        """Show the layer with the given name.
        Args:
            name (str): Layer name
        """
        self.hidden_layers.discard(name)

    def set_design(self, design: QDesign):
        """Set the design.
        Args:
            design (QDesign): The design
        """
        self.design = design
        self.clear_options()
        # TODO

    def clear_options(self):
        """Clear all options."""
        self._hidden_components.clear()
        self.hidden_layers.clear()

    def render(self, ax: Axes):
        """Assumes that the axis has been cleared already and so on.
        Args:
            ax (matplotlib.axes.Axes): mpl axis to draw on
        """

        self.logger.debug('Rendering element tables to plot window.')
        self.render_tables(ax)

    def get_mask(self, table: pd.DataFrame) -> pd.Series:
        """Gets the mask.
        Args:
            table (pd.DataFrame): dataframe
        Returns:
            pd.Series: return pandas index series with boolen mask
            - i.e., which are not hidden or otherwise
        """

        # TODO: Ideally these should be replaced with interface functions,
        # not direct access to underlying internal representation

        mask = table.layer.isin(self.hidden_layers)
        mask = table.component.isin(self._hidden_components)

        return ~mask  # not

    def _render_poly_array(self, ax: Axes, poly_array: np.array, mpl_kw: dict):
        """Render the poly array.
        Args:
            ax (Axes): The axis
            poly_array (np.array): The poly
            mpl_kw (dict): The parameters dictionary
        """
        if len(poly_array) > 0:
            poly_array = to_poly_patch(poly_array)
            ax.add_collection(PatchCollection(poly_array, **mpl_kw))

    @property
    def qgeometry(self) -> 'QGeometryTables':
        """Return the qgeometry of the design."""
        return self.design.qgeometry

    # TODO: move to some config and user input also make widget
    styles = {
        'path': {
            'base': dict(linewidth=2, alpha=0.5),
            'subtracted':
                dict(),  # linestyle='--', edgecolors='k', color='gray'),
            'non-subtracted': dict()
        },
        'poly': {
            'base': dict(linewidth=1, alpha=0.5, edgecolors='k'),
            'subtracted': dict(linestyle='--', color='gray'),
            'non-subtracted': dict()
        },
        'JJ': {
            'base': dict(linewidth=1, alpha=0.2, edgecolors='k'),
            'subtracted': dict(linestyle='--', color='gray'),
            'non-subtracted': dict()
        }
    }
    """Styles"""

    def get_style(self,
                  element_type: str,
                  subtracted=False,
                  layer=None,
                  extra=None):
        """Get the style.
        Args:
            element_type (str): The type of element.
            subtracted (bool): True to subtract the key.  Defaults to False.
            layer (layer): The layer.  Defaults to None.
            extra (dict): Extra stuff to add.  Defaults to None.
        Return:
            dict: Style dictionary
        """
        # element_type - poly path
        extra = extra or {}

        key = 'subtracted' if subtracted else 'non-subtracted'

        kw = {
            **self.styles[element_type].get('base', {}),
            **self.styles[element_type].get(key, {}),
            **extra
        }

        # TODO: maybe pop keys that are invalid for line etc.
        # we could have a validation flag to validate for specific poly / path

        return kw

    def render_tables(self, ax: Axes):
        """Render the tables.
        Args:
            ax (Axes): The axes
        """
        for element_type, table in self.qgeometry.tables.items():
            # Mask the table
            table = table[self.get_mask(table)]

            # subtracted
            mask = table['subtract'] == True
            render_func = getattr(self, f'render_{element_type}')
            render_func(table[mask], ax, subtracted=True)

            # non-subtracted
            table1 = table[~mask]
            # TODO: do by layer and color
            # self.get_color_num()

            # TODO: Check that the function exists
            render_func = getattr(self, f'render_{element_type}')
            render_func(table1, ax, subtracted=False)

    def render_junction(self,
                        table: pd.DataFrame,
                        ax: Axes,
                        subtracted: bool = False,
                        extra_kw: dict = None):
        """Render a table of junction geometry.
        A junction is basically drawn like a path with finite width and no fillet.
        Args:
            table (DataFrame): Element table
            ax (matplotlib.axes.Axes): Axis to render on
            extra_kw (dict): Style params
        """
        if len(table) > 0:
            mask = (table.width == 0) | table.width.isna()
            table1 = table[~mask]
            if len(table1) > 0:
                table1.geometry = table1[['geometry', 'width']].apply(
                    lambda x: x[0].buffer(distance=float(x[1]) / 2.,
                                          cap_style=CAP_STYLE.flat,
                                          join_style=JOIN_STYLE.mitre,
                                          resolution=int(self.options[
                                              'resolution'])),
                    axis=1)
                kw = self.get_style('JJ', subtracted=subtracted, extra=extra_kw)
                self.render_poly(table1, ax, subtracted=subtracted, extra_kw=kw)
            table1 = table[mask]
            if len(table1) > 0:
                self.logger.warning(
                    'One or more junctions have zero width. Consider changing this.'
                )

    def render_poly(self,
                    table: pd.DataFrame,
                    ax: Axes,
                    subtracted: bool = False,
                    extra_kw: dict = None):
        """Render a table of poly geometry.
        Args:
            table (DataFrame): Element table
            ax (matplotlib.axes.Axes): Axis to render on
            kw (dict): Style params
        """
        if len(table) < 1:
            return

        kw = self.get_style('poly', subtracted=subtracted, extra=extra_kw)
        self._render_poly_array(ax, table.geometry, kw)

    def render_fillet(self, table):
        """Renders fillet path.
        Args:
            table (DataFrame): Table of elements with fillets
        Returns:
            DataFrame table with geometry field updated with a polygon filleted path.
        """
        table['geometry'] = table.apply(self.fillet_path, axis=1)
        return table

    def fillet_path(self, row):
        """Output the filleted path.
        Args:
            row (DataFrame): Row to fillet.
        Returns:
            Polygon of the new filleted path.
        """
        path = row["geometry"].coords
        if len(path) <= 2:  # only start and end points, no need to fillet
            return row["geometry"]
        newpath = np.array([path[0]])

        # Get list of vertices that can't be filleted
        no_fillet = bad_fillet_idxs(path, row["fillet"],
                                    self.design.template_options.PRECISION)

        # Iterate through every three-vertex corner
        for (i, (start, corner, end)) in enumerate(zip(path, path[1:],
                                                       path[2:])):
            if i + 1 in no_fillet:  # don't fillet this corner
                newpath = np.concatenate((newpath, np.array([corner])))
            else:
                fillet = self._calc_fillet(np.array(start), np.array(corner),
                                           np.array(end), row["fillet"],
                                           int(self.options['resolution']))
                if fillet is not False:
                    newpath = np.concatenate((newpath, fillet))
                else:
                    newpath = np.concatenate((newpath, np.array([corner])))
        newpath = np.concatenate((newpath, np.array([end])))
        return LineString(newpath)

    def _calc_fillet(self,
                     vertex_start,
                     vertex_corner,
                     vertex_end,
                     radius,
                     points=16):
        """Returns the filleted path based on the start, corner, and end
        vertices and the fillet radius.
        Args:
            vertex_start (np.ndarray): x-y coordinates of starting vertex.
            vertex_corner (np.ndarray): x-y coordinates of corner vertex.
            vertex_end (np.ndarray): x-y coordinates of end vertex.
            radius (float): Fillet radius.
            points (int): Number of points to draw in the fillet corner.
        """
        # Start, corner, and end vertices must be distinct
        if np.array_equal(vertex_start, vertex_corner) or np.array_equal(
                vertex_end, vertex_corner):
            return False

        # Vectors pointing from corner to start and end vertices, respectively
        # Also calculate their lengths and unit vectors
        sc_vec = vertex_start - vertex_corner
        ec_vec = vertex_end - vertex_corner
        sc_norm = np.linalg.norm(sc_vec)
        ec_norm = np.linalg.norm(ec_vec)
        sc_uvec = sc_vec / sc_norm
        ec_uvec = ec_vec / ec_norm

        # Angle between previous unit vectors
        end_angle = np.arccos(np.dot(sc_uvec, ec_uvec))

        # Start, corner, and end vertices can't be collinear
        if (end_angle == 0) or (end_angle == np.pi):
            return False

        # Fillet circle must be small enough to fit inside corner
        if radius / np.tan(end_angle / 2) > min(sc_norm, ec_norm):
            return False

        # Unit vector pointing from corner vertex to center of fillet circle
        net_uvec = (sc_uvec + ec_uvec) / np.linalg.norm(sc_uvec + ec_uvec)

        # Coordinates of center of fillet circle
        circle_center = vertex_corner + net_uvec * radius / np.sin(
            end_angle / 2)

        # Deltas represent displacement from corner vertex to circle center
        # Midpoint angle from circle center to corner, wrt to horizontal extending from former
        # Note: arctan is fine for angles in range (-pi / 2, pi / 2] but needs extra pi factor otherwise
        delta_x = vertex_corner[0] - circle_center[0]
        delta_y = vertex_corner[1] - circle_center[1]
        if delta_x:
            theta_mid = np.arctan(delta_y / delta_x) + np.pi * int(delta_x < 0)
        else:
            theta_mid = np.pi * ((1 - 2 * int(delta_y < 0)) + int(delta_y < 0))

        # Start and end sweep angles determined relative to midpoint angle
        # Swap them as needed to resolve ambiguity in arctan
        theta_start = theta_mid - (np.pi - end_angle) / 2
        theta_end = theta_mid + (np.pi - end_angle) / 2
        p1 = circle_center + radius * np.array(
            [np.cos(theta_start), np.sin(theta_start)])
        p2 = circle_center + radius * np.array(
            [np.cos(theta_end), np.sin(theta_end)])
        if np.linalg.norm(vertex_start - p2) < np.linalg.norm(vertex_start -
                                                              p1):
            theta_start, theta_end = theta_end, theta_start

        # Populate the fillet corner, skipping the start point since it's already added
        path = np.array([
            circle_center + radius * np.array(
                [np.cos(theta_start), np.sin(theta_start)])
        ])
        for theta in np.linspace(theta_start, theta_end, points)[1:]:
            path = np.concatenate(
                (path,
                 np.array([
                     circle_center + radius *
                     np.array([np.cos(theta), np.sin(theta)])
                 ])))
        return path

    def render_path(self,
                    table: pd.DataFrame,
                    ax: Axes,
                    subtracted: bool = False,
                    extra_kw: dict = None):
        """Render a table of path geometry.
        Args:
            table (DataFrame): Element table
            ax (matplotlib.axes.Axes): Axis to render on
            kw (dict): Style params
        """
        if len(table) < 1:
            return

        # mask for all non zero width paths
        # TODO: could there be a problem with float vs int here?
        mask = (table.width == 0) | table.width.isna()
        # print(f'subtracted={subtracted}\n\n')
        # display(table)
        # display(imask)

        # convert to polys - handle non zero width
        table1 = table[~mask]

        mask2 = (table1.fillet == 0)

        table2 = table1[~mask2]

        for index, row in table2[table2.fillet.notnull()].iterrows():
            table1.loc[index, 'geometry'] = self.fillet_path(row)

        if len(table1) > 0:
            table1.geometry = table1[['geometry', 'width']].apply(lambda x: x[
                0].buffer(distance=float(x[1]) / 2.,
                          cap_style=CAP_STYLE.flat,
                          join_style=JOIN_STYLE.mitre,
                          resolution=int(self.options['resolution'])),
                                                                  axis=1)

            kw = self.get_style('poly', subtracted=subtracted, extra=extra_kw)

            # render components
            self.render_poly(table1, ax, subtracted=subtracted, extra_kw=kw)

        # handle zero width
        table1 = table[mask]
        # best way to plot?
        # TODO: speed and vectorize?
        if len(table1) > 0:
            kw = self.get_style('path', subtracted=subtracted, extra=extra_kw)
            line_segments = LineCollection(table1.geometry)
            ax.add_collection(line_segments)


# DEFAULT['renderer_mpl'] = Dict(
#     annot_conectors=Dict(
#         ofst=[0.025, 0.025],
#         annotate_kw=dict(  # called by ax.annotate
#             color='r',
#             arrowprops=dict(color='r', shrink=0.1, width=0.05, headwidth=0.1)
#         ),
#         line_kw=dict(lw=2, c='r')
#     ),
# )

# class QRendererMPL(QRendererGui):
#     """
#     Renderer for matplotlib in a GUI environment.

#     TODO: How do we handle component selection, etc.
#     """
#     name = 'mpl'
#     element_extensions = dict()

#     def render_shapely(self, obj, kw=None):
#         # TODO: simplify, specialize, and update this function
#         # right now, this is just calling the V0.1 old style
#         render(obj, ax=self.ax, kw= {} or kw)

#     def render_connectors(self):
#         '''
#         Plots all connectors on the active axes. Draws the 1D line that
#         represents the "port" of a connector point. These are referenced for smart placement
#             of Metal components, such as when using functions like Metal_CPW_Connect.

#         TODO: add some filter for sense of what components are visible?
#               or on what chip the connectors are
#         '''

#         for name, conn in self.design.connectors.items():

#             line = LineString(conn.points)

#             self.render_shapely(line, kw=DEFAULT.annot_conectors.line_kw)

#             self.ax.annotate(name, xy=conn.middle[:2], xytext=conn.middle +
#                              np.array(DEFAULT.annot_conectors.ofst),
#                              **DEFAULT.annot_conectors.annotate_kw)



from typing import List, Tuple
from ..designs.design_base import QDesign


def get_min_bounding_box(design: QDesign, qcomp_ids: List[int], case: int,
                         logger) -> Tuple[float]:
    """
    Determine the max/min x/y coordinates of the smallest rectangular, axis-aligned
    bounding box that will enclose a selection of components to render, given by
    self.qcomp_ids. This method is only used when box_plus_buffer is True.

    Returns:
        Tuple[float]: min x, min y, max x, and max y coordinates of bounding box.
    """
    min_x_main = min_y_main = float("inf")
    max_x_main = max_y_main = float("-inf")
    if case == 2:  # One or more components not in QDesign.
        logger.warning("One or more components not found.")
    elif case == 1:  # All components rendered.
        for qcomp in design.components:
            min_x, min_y, max_x, max_y = design.components[
                qcomp].qgeometry_bounds()
            min_x_main = min(min_x, min_x_main)
            min_y_main = min(min_y, min_y_main)
            max_x_main = max(max_x, max_x_main)
            max_y_main = max(max_y, max_y_main)
    else:  # Strict subset rendered.
        for qcomp_id in qcomp_ids:
            min_x, min_y, max_x, max_y = design._components[
                qcomp_id].qgeometry_bounds()
            min_x_main = min(min_x, min_x_main)
            min_y_main = min(min_y, min_y_main)
            max_x_main = max(max_x, max_x_main)
            max_y_main = max(max_y, max_y_main)
    return min_x_main, min_y_main, max_x_main, max_y_main
